using System;
using System.IO;
using System.Linq;

namespace Logic
{
  class Program
  {
    static Program()
    {
      EmbeddedResourceLoader.InstallAssemblyResolution();
    }

    /// <summary>
    /// Each command-line argument is interpreted as a path to a file containing a single proposition or argument.  Each file
    /// is parsed and its contents are tested for validity, contingency and inconsistency.
    /// </summary>
    /// <param name="aCommandLineArguments">an array of paths to files containing propositions or arguments</param>
    public static void Main( string[] aCommandLineArguments )
    {
      if ( aCommandLineArguments.Length == 0
          || aCommandLineArguments.Intersects( Utility.MakeArray( "-h", "--help", "/?" ) ) )
      {
        PrintUsage();
      }
      else
      {
        foreach ( string lPath in aCommandLineArguments )
        {
          ProcessFile( lPath );
        }
        
        #if DEBUG
        Console.ReadLine();
        #endif
      }
    }

    private static void PrintUsage()
    {
      Console.WriteLine( "Usage: mpl.exe [proposition file] [another proposition file] ..." );
      Console.WriteLine();
      Console.Write( "Parse each file in the command line and decide if the proposition in the file is " );
      Console.WriteLine( "necessarily true, contingent, or self-contradictory." );
    }

    private static void ProcessFile( string aPath )
    {
      if ( !File.Exists( aPath ) )
      {
        Console.WriteLine( "File not found: {0}", aPath );
        return;
      }
      
      try
      {
        string[] lFileContents = File.ReadAllLines( aPath );
        Matrix lProposition = Parser.Parse( lFileContents );
        if ( lFileContents.Any( fLine => fLine.Trim() == "->" ) )
        {
          Console.Write( "The argument in {0} is ", aPath );
          if ( Factory.Necessarily( lProposition ).TrueInMaximallyDiverseUniverse )
            Console.WriteLine( "valid." );
          else
            Console.WriteLine( "invalid." );
        }
        else
        {
          Console.Write( "The contents of {0} are ", aPath );
          if ( Factory.Necessarily( lProposition ).TrueInMaximallyDiverseUniverse )
            Console.WriteLine( "necessarily true." );
          else if ( Factory.Possibly( lProposition ).TrueInMaximallyDiverseUniverse )
            Console.WriteLine( "possible." );
          else
            Console.WriteLine( "self-contradictory." );
        }
      }
      catch ( Exception lException )
      {
        Console.WriteLine( "invalid:" );
        Console.WriteLine( "{0}: {1}", lException.GetType(), lException.Message );
      }
    }
  }
}
