﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;

namespace Logic
{
  /// <summary>
  /// Contains methods for parsing symbolic logic expressions.
  /// </summary>
  public static class Parser
  {
    /// <summary>
    /// Parse a file containing a list of statements or a logical argument.  Ignore comments and whitespace.
    /// Interpret each line as a statement; conjoin the statements into one statement.  If one line contains
    /// a binary operator, conjoin all statements before it and make them the left side of the operation and
    /// conjoin all statements after it and make them the right side of the operation.  Bind all free
    /// variables in existential quantifiers.
    /// </summary>
    /// <param name="aFileContents"></param>
    /// <returns></returns>
    public static Matrix Parse( string[] aFileContents )
    {
      List<string> lAntecedents = new List<string>();
      List<string> lConsequents = new List<string>();
      string lConnective = null;
      foreach ( string line in aFileContents )
      {
        // Get rid of comments.
        string lAdjustedLine = Regex.Replace( line, @"//.*$", "" );

        // Get rid of whitespace.
        lAdjustedLine = Regex.Replace( lAdjustedLine, @"\s+", "" );

        // Skip empty lines.
        if ( lAdjustedLine.Length == 0 )
          continue;

        if ( lAdjustedLine.IsBinaryOperator() )
        {
          lConnective = lAdjustedLine;
          continue;
        }

        if ( lConnective == null )
          lAntecedents.Add( lAdjustedLine );
        else
          lConsequents.Add( lAdjustedLine );
      }

      if ( lConnective != null )
      {
        if ( lAntecedents.Count == 0 )
          throw new ParseError( "No antecedents found in argument" );
        else if ( lConsequents.Count == 0 )
          throw new ParseError( "No consequents found in argument" );
        else
          return Parse( Parenthesize( Conjoin( lAntecedents ) ) + lConnective + Parenthesize( Conjoin( lConsequents ) ) );
      }
      else
      {
        if ( lAntecedents.Count == 0 )
          throw new ParseError( "No content found." );
        else
          return Parse( Conjoin( lAntecedents ) );
      }
    }

    private static Matrix Parse( string aString )
    {
      PredicateDictionary lPredicateDictionary = new PredicateDictionary();
      VariableDictionary lVariableDictionary = new VariableDictionary();
      Matrix lParsedMatrix;
      List<char> lFreeVariableSymbols = new List<char>();

      while ( true )
      {
        try
        {
          lParsedMatrix = Parse( aString.GetSubexpressions(), lPredicateDictionary, lVariableDictionary );
          break;
        }
        catch ( FreeVariableNotFoundException lException )
        {
          lFreeVariableSymbols.Add( lException.VariableSymbol );
          lVariableDictionary = lVariableDictionary.UpdateWith( lException.VariableSymbol, Factory.Variable( lException.VariableSymbol ) );
        }
      }

      foreach ( char lSymbol in lFreeVariableSymbols )
      {
        lParsedMatrix = Factory.ThereExists( lVariableDictionary.Retrieve( lSymbol ), lParsedMatrix );
      }

      return lParsedMatrix;
    }
   

    private static Regex ExactMatch( string aPattern )
    {
      return new Regex( string.Format( "^{0}$", aPattern ) );
    }
    private static Regex ThereExists = ExactMatch( @"3[a-z]," );
    private static Regex ForAll = ExactMatch( @"[a-z]," );
    private static Regex And = ExactMatch( @"&" );
    private static Regex Or = ExactMatch( @"\|" );
    private static Regex OnlyIf = ExactMatch( @"->" );
    private static Regex IfAndOnlyIf = ExactMatch( @"<=>" );
    private static Regex Not = ExactMatch( @"~" );
    private static Regex Nor = ExactMatch( @"!" );
    private static Regex Is = ExactMatch( @"[A-Z][a-z]" );
    private static Regex Possibly = ExactMatch( @"<>" );
    private static Regex Necessarily = ExactMatch( @"\[\]" );
    private static Regex ParenthesizedExpression = ExactMatch( @"\(.*\)" );

    private static IEnumerable<IEnumerable<T>> Chomble<T>( IEnumerable<T> aString, Predicate<IEnumerable<T>> aPredicate )
    {
      Queue<T> lAccumulation = new Queue<T>();

      foreach ( T lItem in aString )
      {
        lAccumulation.Enqueue( lItem );

        if ( aPredicate( lAccumulation ) )
        {
          yield return lAccumulation.ToArray();
          lAccumulation.Clear();
        }
      }
    }

    private static char GetSymbolForVariable( string aQuantifier )
    {
      if ( aQuantifier.Matches( ThereExists ) )
        return aQuantifier[ 1 ];
      else if ( aQuantifier.Matches( ForAll ) )
        return aQuantifier[ 0 ];
      else
        throw new ApplicationException();
    }

    private static T[] Array<T>( params T[] aArray )
    {
      return aArray;
    }

    private static bool IsAnyOf( this string aString, params Regex[] aRegularExpressions )
    {
      return aRegularExpressions.Any( fRegex => fRegex.IsMatch( aString ) );
    }

    private static bool IsParenthesizedGroup( this string aString )
    {
      return aString.StartsWith( "(" ) && aString.HasMatchingParentheses();
    }

    private static bool IsToken( this string aString )
    {
      return aString.IsParenthesizedGroup() || aString.IsAnyOf(
        And,
        Or,
        IfAndOnlyIf,
        ThereExists,
        OnlyIf,
        Not,
        Nor,
        Is,
        ForAll,
        Possibly,
        Necessarily );
    }

    private static IEnumerable<string> GetSubexpressions( this string aString )
    {
      return Chomble( aString, fAccumulation => fAccumulation.IsToken() );
    }

    private static bool IsBinaryOperator( this string aString )
    {
      return aString.IsAnyOf( And, Or, IfAndOnlyIf, OnlyIf, Nor );
    }

    private static T Second<T>( this IEnumerable<T> aEnumerable )
    {
      return aEnumerable.Skip( 1 ).First();
    }

    private static IEnumerable<T> UpToLast<T>( this IEnumerable<T> aEnumerable, Predicate<T> aPredicate )
    {
      return aEnumerable.Reverse().AfterLast( fItem => aPredicate( fItem ) ).Reverse();
    }

    private static IEnumerable<T> AfterLast<T>( this IEnumerable<T> aEnumerable, Predicate<T> aPredicate )
    {
      return aEnumerable.SkipWhile( fItem => !aPredicate( fItem ) ).Skip( 1 );
    } 

    private static string Conjoin( IEnumerable<string> aExpressions )
    {
      return string.Join( "&", aExpressions.Select( fExpression => Parenthesize( fExpression ) ).ToArray() );
    }

    private static string Parenthesize( string aExpression )
    {
      return string.Format( "({0})", aExpression );
    }

    private static Matrix ParsePredicationOrGrouping(
      string aExpression,
      PredicateDictionary aCollectedPredicates,
      VariableDictionary aFreeVariables )
    {
      if ( aExpression.Matches( Is ) )
      {
        return Factory.Predication(
          aCollectedPredicates.Add( aExpression[ 0 ] ),
          aFreeVariables.Retrieve( aExpression[ 1 ] ) );
      }

      if ( aExpression.Matches( ParenthesizedExpression ) )
      {
        return Parse(
          aExpression.Substring( 1, aExpression.Length - 2 ).GetSubexpressions(),
          aCollectedPredicates,
          aFreeVariables );
      }

      throw new ParseError( "Expected predication or parenthesized expression, found \"{0}\"", aExpression );
    }

    private static Matrix ParseQuantification(
      string aQuantifier,
      IEnumerable<string> aBody,
      PredicateDictionary aCollectedPredicates,
      VariableDictionary aFreeVariables )
    {
      char lSymbolForVariable = GetSymbolForVariable( aQuantifier );
      Variable lVariable = Factory.Variable( lSymbolForVariable );
      Matrix lBody = Parse(
        aBody,
        aCollectedPredicates,
        aFreeVariables.UpdateWith( lSymbolForVariable, lVariable ) );

      if ( aQuantifier.Matches( ForAll ) )
        return Factory.ForAll( lVariable, lBody );

      if ( aQuantifier.Matches( ThereExists ) )
        return Factory.ThereExists( lVariable, lBody );

      throw new ApplicationException();
    }

    private static Matrix ParseBinaryOperator(
      IEnumerable<string> aTokens,
      PredicateDictionary aCollectedPredicates,
      VariableDictionary aFreeVariables )
    {
      IEnumerable<string> lTokensBeforeLastBinaryOperator = aTokens.UpToLast( fToken => fToken.IsBinaryOperator() );
      IEnumerable<string> lTokensAfterLastBinaryOperator = aTokens.AfterLast( fToken => fToken.IsBinaryOperator() );
      string lLastBinaryOperator = aTokens.Last( fToken => fToken.IsBinaryOperator() );

      Matrix lLeft = Parse( lTokensBeforeLastBinaryOperator, aCollectedPredicates, aFreeVariables );
      Matrix lRight = Parse( lTokensAfterLastBinaryOperator, aCollectedPredicates, aFreeVariables );

      if ( lLastBinaryOperator.Matches( And ) )
        return Factory.And( lLeft, lRight );

      if ( lLastBinaryOperator.Matches( Or ) )
        return Factory.Or( lLeft, lRight );

      if ( lLastBinaryOperator.Matches( OnlyIf ) )
        return Factory.OnlyIf( lLeft, lRight );

      if ( lLastBinaryOperator.Matches( IfAndOnlyIf ) )
        return Factory.IfAndOnlyIf( lLeft, lRight );

      if ( lLastBinaryOperator.Matches( Nor ) )
        return Factory.Nor( lLeft, lRight );

      throw new ApplicationException();
    }

    private static Matrix Parse(
      IEnumerable<string> aExpressions,
      PredicateDictionary aCollectedPredicates,
      VariableDictionary aFreeVariables )
    {
      string lFirst = aExpressions.First();

      if ( aExpressions.Count() == 1 )
        return ParsePredicationOrGrouping( lFirst, aCollectedPredicates, aFreeVariables );

      string lSecond = aExpressions.Second();
      IEnumerable<string> lTail = aExpressions.Tail();

      if ( AreUnariesFollowedByAQuantifier( aExpressions ) )
        return ParseUnaryOperatorOnTail( lFirst, lTail, aCollectedPredicates, aFreeVariables );

      if ( lFirst.Matches( ForAll ) || lFirst.Matches( ThereExists ) )
        return ParseQuantification( lFirst, lTail, aCollectedPredicates, aFreeVariables );

      if ( aExpressions.Any( fToken => fToken.IsBinaryOperator() ) )
        return ParseBinaryOperator( aExpressions, aCollectedPredicates, aFreeVariables );

      if ( IsUnaryOperator( lFirst ) )
        return ParseUnaryOperatorOnTail( lFirst, lTail, aCollectedPredicates, aFreeVariables );

      throw new ApplicationException();
    }
    
    private static bool IsUnaryOperator( string aExpression )
    {
      return Not.IsMatch( aExpression ) || Possibly.IsMatch( aExpression ) || Necessarily.IsMatch( aExpression );
    }
    
    private static bool AreUnariesFollowedByAQuantifier( IEnumerable<string> aExpressions )
    {
      if ( IsUnaryOperator( aExpressions.First() ) )
        return aExpressions.SkipWhile( fToken => IsUnaryOperator( fToken ) ).First().IsAnyOf( ForAll, ThereExists );
      else
        return false;
    }
    
    private static Matrix ParseUnaryOperatorOnTail(
      string aUnaryOperator,
      IEnumerable<string> aTail,
      PredicateDictionary aCollectedPredicates,
      VariableDictionary aFreeVariables )
    {
      Matrix lTail = Parse( aTail, aCollectedPredicates, aFreeVariables );
      
      if ( aUnaryOperator.Matches( Not ) )
        return Factory.Not( lTail );
        
      if ( aUnaryOperator.Matches( Possibly ) )
        return Factory.Possibly( lTail );
        
      if ( aUnaryOperator.Matches( Necessarily ) )
        return Factory.Necessarily( lTail );
        
      throw new ApplicationException();
    }

    private static IEnumerable<string> Chomble( string aString, Predicate<string> aPredicate )
    {
      return Chomble( aString.ToCharArray(), lSubstring => aPredicate( new string( lSubstring.ToArray() ) ) )
        .Select( lCharacters => new string( lCharacters.ToArray() ) );
    }

    private static bool HasMatchingParentheses( this string aString )
    {
      int lCount = 0;
      bool lHasParentheses = false;
      foreach ( char lChar in aString )
      {
        switch ( lChar )
        {
          case '(':
            lCount++;
            lHasParentheses = true;
            break;
          case ')':
            lCount--;
            break;
        }

        if ( lCount < 0 )
          return false;

      }
      return lHasParentheses && lCount == 0;
    }
    
    private static bool Matches( this string aString, Regex aRegularExpression )
    {
      return aRegularExpression.IsMatch( aString );
    }
  }
}
